#
# Copyright (c) Daniel Sheffield 2021 - 2022
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from psycopg.sql import (
    Identifier,
    SQL,
    Literal,
    Placeholder,
    Composed,
)
from collections import (
    OrderedDict,
)

def get_where(unit, product=None, category=None, group=None, organic=None, limit='90 days'):
    where = [ ]
    if product is not None:
        where.append(SQL(' ').join([
            Identifier('products', 'name'),
            SQL('='),
            Literal(product)
        ]))
    if category is not None:
        where.append(SQL(' ').join([
            Identifier('categories', 'name'),
            SQL('='),
            Literal(category)
        ]))
    if group is not None:
        where.append(SQL(' ').join([
            Identifier('groups', 'name'),
            SQL('='),
            Literal(group)
        ]))
    if organic is not None:
        where.append(SQL(' ').join([
            Identifier('organic'),
            SQL('='),
            Literal(organic),
        ]))
    if limit:
        where.append(
            SQL("{ts} at time zone 'utc' BETWEEN now() - {interval} AND now()").format(
                ts=Identifier('ts'),
                interval=SQL("{literal}::interval").format(literal=Literal(limit))
            )
        )
    where.append(
        SQL("(SELECT id FROM units WHERE name = {unit}) = conv._to").format(
            unit=Literal(unit),
        )
    )
    return SQL('').join([
        SQL("WHERE"
            "\n      "),
        SQL("\n  AND ").join(where),
    ])

def get_historic_prices_statement(unit, sort=None, product=None, category=None, group=None, organic=None, limit='90 days'):
    partition = f"(PARTITION BY {'organic,' if organic is not None else ''} product_id)"
    organic_sort = f"{'organic,' if organic is not None else ''}"
    sort_sql = SQL('').join([
        SQL('{sort} {direction},').format(
            sort=Identifier(f'{sort}'),
            direction = SQL('DESC' if sort == 'ts' else 'ASC')
        ),
    ]) if sort is not None else SQL('')

    _with = SQL("""WITH conv AS (
    SELECT
        _1._from, _1._to, factor FROM conversions AS _1
    UNION
    SELECT _2._to, _2._from, 1/_2.factor FROM conversions AS _2
    UNION
    SELECT _3._from, _3._from, 1 FROM conversions AS _3
    UNION
    SELECT _4.id AS _from, _4.id AS _to, 1 FROM units AS _4
)""")
    select = OrderedDict([
        ('id', Identifier('transactions','id')),
        ('ts_raw', SQL("""(transactions.ts AT TIME ZONE 'UTC')::timestamp without time zone""")),
        ('%d/%m/%y %_I%P', SQL("""(transactions.ts AT TIME ZONE 'UTC')::timestamp without time zone""")),
        ('code', Identifier('stores', 'code')),
        ('$/unit', SQL("""TRUNC(price / (quantity * factor), 4)""")),
        ('avg', SQL(f"""TRUNC(sum(price) OVER {partition} / sum(quantity * factor) OVER {partition}, 4)""")),
        ('min', SQL(f"""TRUNC(min(price / (quantity * factor)) OVER {partition}, 4)""")),
        ('max', SQL(f"""TRUNC(max(price / (quantity * factor)) OVER {partition}, 4)""")),
        ('price', SQL(f"""TRUNC(price, 4)::numeric""")),
        ('quantity', SQL(f"""TRUNC(quantity*factor, 4)::numeric""")),
        ('product', Identifier('products','name')),
        ('category', Identifier('categories', 'name')),
        ('group', Identifier('groups', 'name')),
        ('organic', Identifier('organic')),
    ])
    statement = SQL('\n').join([
        _with,
        SQL('').join([
            SQL("SELECT"
                "\n  "),
            SQL(','
                "\n  ").join([
                SQL(' ').join([v, SQL('AS'), Identifier(f'{k}')]) for k, v in select.items()
            ])
        ]),
        SQL('').join([
            SQL("FROM"
                "\n       "),
            SQL("\n  JOIN ").join([
                SQL("transactions"),
                SQL("{table} ON ({table}.{key} = {index})").format(
                    table=Identifier('conv'),
                    key=Identifier('_from'),
                    index=Identifier('unit_id'),
                ),
                SQL("{table} ON {table}.{key} = {index}").format(
                    table=Identifier('units'),
                    key=Identifier('id'),
                    index=Identifier('unit_id'),
                ),
                SQL("{table} ON {table}.{key} = {index}").format(
                    table=Identifier('stores'),
                    key=Identifier('id'),
                    index=Identifier('store_id')
                ),
                SQL("{table} ON {table}.{key} = {index}").format(
                    table=Identifier('products'),
                    key=Identifier('id'),
                    index=Identifier('product_id')
                ),
                SQL("{table} ON {table}.{key} = {index}").format(
                    table=Identifier('categories'),
                    key=Identifier('id'),
                    index=Identifier('category_id')
                ),
                SQL("{table} ON {table}.{key} = {index}").format(
                    table=Identifier('groups'),
                    key=Identifier('id'),
                    index=Identifier('group_id')
                ),
            ]),
        ]),
        get_where(unit, product=product, category=category, group=group, organic=organic, limit=limit),
        SQL('ORDER BY {organic_sort} {sort} code, product, category, "group", "$/unit" ASC, ts DESC').format(
            sort=sort_sql,
            organic_sort=SQL(organic_sort),
        ),
    ])
    return statement
